﻿using Microsoft.Extensions.Logging;

namespace Bmp.Core.Common.EventBus;

public abstract class InProcessEventBusBase : IEventBus
{
    private readonly ILogger<InProcessEventBusBase> _logger;

    private readonly Dictionary<Type, HashSet<Delegate>> _dicTypeToHandlers = new();

    protected InProcessEventBusBase(ILogger<InProcessEventBusBase> logger)
    {
        _logger = logger;
    }

    public bool Subscript<T>(Action<T> callBack)
    {
        var type = typeof(T);
        lock (_dicTypeToHandlers)
        {
            if (!_dicTypeToHandlers.TryGetValue(type, out var handlers))
            {
                handlers = _dicTypeToHandlers[type] = new();
            }

            return handlers.Add(callBack); // 忽略重复
        }
    }

    public bool UnSubscript<T>(Action<T> callBack)
    {
        lock (_dicTypeToHandlers)
        {
            if (_dicTypeToHandlers.TryGetValue(typeof(T), out var handlers))
            {
                var unSubscript = handlers.Remove(callBack);

                if (handlers.Count == 0) _dicTypeToHandlers.Remove(typeof(T));

                return unSubscript;
            }

            return false;
        }
    }

    public bool Publish<T>()
    {
        PublishInternal(new AnyPublishEvent(typeof(T), default));
        return PublishInternal<T>(default);
    }

    public bool Publish<T>(T obj)
    {
        PublishInternal(new AnyPublishEvent(typeof(T), obj));
        return PublishInternal(obj);
    }

    private bool PublishInternal<T>(T eventValue)
    {
        var type = typeof(T);

        Delegate[] subscripts;
        lock (_dicTypeToHandlers)
        {
            if (!_dicTypeToHandlers.TryGetValue(type, out var handlers)) return false;
            subscripts = handlers.ToArray();
        }

        foreach (var del in subscripts)
        {
            try
            {
                ((Action<T>)del)(eventValue);
            }
            catch (Exception e)
            {
                _logger.LogError(e, nameof(Publish));
            }
        }

        return true;
    }
}
