﻿using Bmp.Core.Lite.Playback.Outputs.NAudioASIO.Originals;
using NAudio.Wave;

namespace Bmp.Core.Lite.Playback.Outputs.NAudioASIO;

/// <summary>
/// This class stores convertors for different interleaved WaveFormat to ASIOSampleType separate channel
/// format.
/// </summary>
internal class AsioSampleConvertor
{
    public delegate void SampleConvertor(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples);

    /// <summary>
    /// Selects the sample convertor based on the input WaveFormat and the output ASIOSampleTtype.
    /// </summary>
    /// <param name="waveFormat">The wave format.</param>
    /// <param name="asioType">The type.</param>
    /// <param name="bitPerRawSample"></param>
    /// <param name="asioSampleType"></param>
    /// <returns></returns>
    public static SampleConvertor SelectSampleConvertor(WaveFormat waveFormat, AsioSampleType asioType, int? bitPerRawSample, AsioSampleType? asioSampleType)
    {
        SampleConvertor convertor = null;
        var is2Channels = waveFormat.Channels == 2;

        // TODO: AsioSampleConvertor IMPLEMENTS OTHER CONVERTOR TYPES
        switch (asioType)
        {
            case AsioSampleType.DSDInt8LSB1:
                if (asioSampleType == AsioSampleType.DSDInt8LSB1)
                {
                    convertor = is2Channels ? ConvertedBytePerChannelStereoDsdCopy : ConvertedBytePerChannelMonoDsdCopy;
                }
                else if (asioSampleType == AsioSampleType.DSDInt8MSB1)
                {
                    convertor = is2Channels ? ConvertedBytePerChannelStereoDsdRsv : ConvertedBytePerChannelMonoDsdRsv;
                }
                break;

            case AsioSampleType.DSDInt8MSB1:
                if (asioSampleType == AsioSampleType.DSDInt8MSB1)
                {
                    convertor = is2Channels ? ConvertedBytePerChannelStereoDsdCopy : ConvertedBytePerChannelMonoDsdCopy;
                }
                else if (asioSampleType == AsioSampleType.DSDInt8LSB1)
                {
                    convertor = is2Channels ? ConvertedBytePerChannelStereoDsdRsv : ConvertedBytePerChannelMonoDsdRsv;
                }
                break;

            case AsioSampleType.Int32LSB:
                switch (waveFormat.BitsPerSample)
                {
                    case 16:
                        convertor = is2Channels ? ConvertorShortToInt2Channels : ConvertorShortToIntGeneric;
                        break;

                    case 32:
                        if (waveFormat.Encoding == WaveFormatEncoding.IeeeFloat)
                            convertor = is2Channels ? ConvertorFloatToInt2Channels : ConvertorFloatToIntGeneric;
                        else
                            convertor = is2Channels ? ConvertorIntToInt2Channels : ConvertorIntToIntGeneric;
                        break;
                }
                break;

            case AsioSampleType.Int16LSB:
                switch (waveFormat.BitsPerSample)
                {
                    case 16:
                        convertor = is2Channels ? ConvertorShortToShort2Channels : ConvertorShortToShortGeneric;
                        break;

                    case 32:
                        if (waveFormat.Encoding == WaveFormatEncoding.IeeeFloat)
                            convertor = is2Channels ? ConvertorFloatToShort2Channels : ConvertorFloatToShortGeneric;
                        else
                            convertor = is2Channels ? ConvertorIntToShort2Channels : ConvertorIntToShortGeneric;
                        break;
                }
                break;

            case AsioSampleType.Int24LSB:
                switch (waveFormat.BitsPerSample)
                {
                    case 16:
                        if (waveFormat.Encoding == WaveFormatEncoding.Pcm)
                        {
                            convertor = PaddingPcm16To24; // BmpMod: PaddingPcm16To24
                        }
                        else throw new ArgumentException("Not a supported conversion");
                        break;

                    case 24:
                        if (waveFormat.Encoding == WaveFormatEncoding.Pcm)
                        {
                            convertor = ConverterPcm24To24;
                        }
                        else throw new ArgumentException("Not a supported conversion");
                        break;

                    case 32:
                        if (waveFormat.Encoding == WaveFormatEncoding.IeeeFloat)
                            convertor = ConverterFloatTo24LsbGeneric;
                        else if (waveFormat.Encoding == WaveFormatEncoding.Pcm && bitPerRawSample == 24)
                            convertor = TruncatePcm32To24; // BmpMod: TruncatePcm32To24
                        else
                            throw new ArgumentException("Not a supported conversion");
                        break;
                }
                break;

            case AsioSampleType.Float32LSB:
                switch (waveFormat.BitsPerSample)
                {
                    case 16:
                        throw new ArgumentException("Not a supported conversion");
                    case 32:
                        if (waveFormat.Encoding == WaveFormatEncoding.IeeeFloat)
                            convertor = ConverterFloatToFloatGeneric;
                        else
                            convertor = ConvertorIntToFloatGeneric;
                        break;
                }
                break;
        }

        if (convertor == null) throw new ArgumentException($"ASIO Buffer Type {Enum.GetName(typeof(AsioSampleType), asioType)} is not yet supported.");

        return convertor;
    }

    /// <summary>
    /// Optimized convertor for 2 channels SHORT
    /// </summary>
    public static void ConvertorShortToInt2Channels(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (short*)inputInterleavedBuffer;
            // Use a trick (short instead of int to avoid any convertion from 16Bit to 32Bit)
            var leftSamples = (short*)asioOutputBuffers[0];
            var rightSamples = (short*)asioOutputBuffers[1];

            // Point to upper 16 bits of the 32Bits.
            leftSamples++;
            rightSamples++;
            for (var i = 0; i < nbSamples; i++)
            {
                *leftSamples = inputSamples[0];
                *rightSamples = inputSamples[1];
                // Go to next sample
                inputSamples += 2;
                // Add 4 Bytes
                leftSamples += 2;
                rightSamples += 2;
            }
        }
    }

    /// <summary>
    /// Generic convertor for SHORT
    /// </summary>
    public static void ConvertorShortToIntGeneric(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (short*)inputInterleavedBuffer;
            // Use a trick (short instead of int to avoid any convertion from 16Bit to 32Bit)
            var samples = new short*[nbChannels];
            for (var i = 0; i < nbChannels; i++)
            {
                samples[i] = (short*)asioOutputBuffers[i];
                // Point to upper 16 bits of the 32Bits.
                samples[i]++;
            }

            for (var i = 0; i < nbSamples; i++)
            {
                for (var j = 0; j < nbChannels; j++)
                {
                    *samples[j] = *inputSamples++;
                    samples[j] += 2;
                }
            }
        }
    }

    /// <summary>
    /// Optimized convertor for 2 channels FLOAT
    /// </summary>
    public static void ConvertorFloatToInt2Channels(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (float*)inputInterleavedBuffer;
            var leftSamples = (int*)asioOutputBuffers[0];
            var rightSamples = (int*)asioOutputBuffers[1];

            for (var i = 0; i < nbSamples; i++)
            {
                *leftSamples++ = ClampToInt(inputSamples[0]);
                *rightSamples++ = ClampToInt(inputSamples[1]);
                inputSamples += 2;
            }
        }
    }

    /// <summary>
    /// Generic convertor Float to INT
    /// </summary>
    public static void ConvertorFloatToIntGeneric(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (float*)inputInterleavedBuffer;
            var samples = new int*[nbChannels];
            for (var i = 0; i < nbChannels; i++)
            {
                samples[i] = (int*)asioOutputBuffers[i];
            }

            for (var i = 0; i < nbSamples; i++)
            {
                for (var j = 0; j < nbChannels; j++)
                {
                    *samples[j]++ = ClampToInt(*inputSamples++);
                }
            }
        }
    }

    /// <summary>
    /// Optimized convertor for 2 channels INT to INT
    /// </summary>
    public static void ConvertorIntToInt2Channels(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (int*)inputInterleavedBuffer;
            var leftSamples = (int*)asioOutputBuffers[0];
            var rightSamples = (int*)asioOutputBuffers[1];

            for (var i = 0; i < nbSamples; i++)
            {
                *leftSamples++ = inputSamples[0];
                *rightSamples++ = inputSamples[1];
                inputSamples += 2;
            }
        }
    }

    /// <summary>
    /// Generic convertor INT to INT
    /// </summary>
    public static void ConvertorIntToIntGeneric(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (int*)inputInterleavedBuffer;
            var samples = new int*[nbChannels];
            for (var i = 0; i < nbChannels; i++)
            {
                samples[i] = (int*)asioOutputBuffers[i];
            }

            for (var i = 0; i < nbSamples; i++)
            {
                for (var j = 0; j < nbChannels; j++)
                {
                    *samples[j]++ = *inputSamples++;
                }
            }
        }
    }

    /// <summary>
    /// Optimized convertor for 2 channels INT to SHORT
    /// </summary>
    public static void ConvertorIntToShort2Channels(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (int*)inputInterleavedBuffer;
            var leftSamples = (short*)asioOutputBuffers[0];
            var rightSamples = (short*)asioOutputBuffers[1];

            for (var i = 0; i < nbSamples; i++)
            {
                *leftSamples++ = (short)(inputSamples![0] / (1 << 16));
                *rightSamples++ = (short)(inputSamples[1] / (1 << 16));
                inputSamples += 2;
            }
        }
    }

    /// <summary>
    /// Generic convertor INT to SHORT
    /// </summary>
    public static void ConvertorIntToShortGeneric(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (int*)inputInterleavedBuffer;
            var samples = new int*[nbChannels];
            for (var i = 0; i < nbChannels; i++)
            {
                samples[i] = (int*)asioOutputBuffers[i];
            }

            for (var i = 0; i < nbSamples; i++)
            {
                for (var j = 0; j < nbChannels; j++)
                {
                    *samples[j]++ = (short)(*inputSamples++ / (1 << 16));
                }
            }
        }
    }

    /// <summary>
    /// Generic convertor INT to FLOAT
    /// </summary>
    public static void ConvertorIntToFloatGeneric(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (int*)inputInterleavedBuffer;
            var samples = new float*[nbChannels];
            for (var i = 0; i < nbChannels; i++)
            {
                samples[i] = (float*)asioOutputBuffers[i];
            }

            for (var i = 0; i < nbSamples; i++)
            {
                for (var j = 0; j < nbChannels; j++)
                {
                    *samples[j]++ = *inputSamples++ / (1 << 32 - 1);
                }
            }
        }
    }

    /// <summary>
    /// Optimized convertor for 2 channels SHORT
    /// </summary>
    public static void ConvertorShortToShort2Channels(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (short*)inputInterleavedBuffer;
            // Use a trick (short instead of int to avoid any convertion from 16Bit to 32Bit)
            var leftSamples = (short*)asioOutputBuffers[0];
            var rightSamples = (short*)asioOutputBuffers[1];

            // Point to upper 16 bits of the 32Bits.
            for (var i = 0; i < nbSamples; i++)
            {
                *leftSamples++ = inputSamples[0];
                *rightSamples++ = inputSamples[1];
                // Go to next sample
                inputSamples += 2;
            }
        }
    }

    /// <summary>
    /// Generic convertor for SHORT
    /// </summary>
    public static void ConvertorShortToShortGeneric(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (short*)inputInterleavedBuffer;
            // Use a trick (short instead of int to avoid any convertion from 16Bit to 32Bit)
            var samples = new short*[nbChannels];
            for (var i = 0; i < nbChannels; i++)
            {
                samples[i] = (short*)asioOutputBuffers[i];
            }

            for (var i = 0; i < nbSamples; i++)
            {
                for (var j = 0; j < nbChannels; j++)
                {
                    *samples[j]++ = *inputSamples++;
                }
            }
        }
    }

    /// <summary>
    /// Optimized convertor for 2 channels FLOAT
    /// </summary>
    public static void ConvertorFloatToShort2Channels(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (float*)inputInterleavedBuffer;
            // Use a trick (short instead of int to avoid any convertion from 16Bit to 32Bit)
            var leftSamples = (short*)asioOutputBuffers[0];
            var rightSamples = (short*)asioOutputBuffers[1];

            for (var i = 0; i < nbSamples; i++)
            {
                *leftSamples++ = ClampToShort(inputSamples[0]);
                *rightSamples++ = ClampToShort(inputSamples[1]);
                inputSamples += 2;
            }
        }
    }

    /// <summary>
    /// Generic convertor SHORT
    /// </summary>
    public static void ConvertorFloatToShortGeneric(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (float*)inputInterleavedBuffer;
            // Use a trick (short instead of int to avoid any convertion from 16Bit to 32Bit)
            var samples = new short*[nbChannels];
            for (var i = 0; i < nbChannels; i++)
            {
                samples[i] = (short*)asioOutputBuffers[i];
            }

            for (var i = 0; i < nbSamples; i++)
            {
                for (var j = 0; j < nbChannels; j++)
                {
                    *samples[j]++ = ClampToShort(*inputSamples++);
                }
            }
        }
    }

    /// <summary>
    /// Generic converter 24 LSB
    /// </summary>
    public static void ConverterFloatTo24LsbGeneric(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (float*)inputInterleavedBuffer;

            var samples = new byte*[nbChannels];
            for (var i = 0; i < nbChannels; i++)
            {
                samples[i] = (byte*)asioOutputBuffers[i];
            }

            for (var i = 0; i < nbSamples; i++)
            {
                for (var j = 0; j < nbChannels; j++)
                {
                    var sample24 = ClampTo24Bit(*inputSamples++);
                    *samples[j]++ = (byte)sample24;
                    *samples[j]++ = (byte)(sample24 >> 8);
                    *samples[j]++ = (byte)(sample24 >> 16);
                }
            }
        }
    }

    /// <summary>
    /// Generic converter 24 LSB
    /// </summary>
    public static void ConverterPcm24To24(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (byte*)inputInterleavedBuffer;

            var pChannel = new byte*[nbChannels];
            for (var i = 0; i < nbChannels; i++)
            {
                pChannel[i] = (byte*)asioOutputBuffers[i];
            }

            for (var i = 0; i < nbSamples; i++)
            {
                for (var ch = 0; ch < nbChannels; ch++)
                {
                    *pChannel[ch]++ = *inputSamples++;
                    *pChannel[ch]++ = *inputSamples++;
                    *pChannel[ch]++ = *inputSamples++;
                }
            }
        }
    }

    /// <summary>
    /// Generic convertor for float
    /// </summary>
    public static void ConverterFloatToFloatGeneric(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (float*)inputInterleavedBuffer;
            var samples = new float*[nbChannels];
            for (var i = 0; i < nbChannels; i++)
            {
                samples[i] = (float*)asioOutputBuffers[i];
            }

            for (var i = 0; i < nbSamples; i++)
            {
                for (var j = 0; j < nbChannels; j++)
                {
                    *samples[j]++ = *inputSamples++;
                }
            }
        }
    }

    // BmpMod: TruncatePcm32To24
    public static void TruncatePcm32To24(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (byte*)inputInterleavedBuffer;

            var pChannel = new byte*[nbChannels];
            for (var i = 0; i < nbChannels; i++) pChannel[i] = (byte*)asioOutputBuffers[i];

            for (var i = 0; i < nbSamples; i++)
            {
                for (var ch = 0; ch < nbChannels; ch++)
                {
                    inputSamples++; // drop byte
                    *pChannel[ch]++ = *inputSamples++;
                    *pChannel[ch]++ = *inputSamples++;
                    *pChannel[ch]++ = *inputSamples++;
                }
            }
        }
    }

    // BmpMod: PaddingPcm16To24
    public static void PaddingPcm16To24(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (byte*)inputInterleavedBuffer;

            var pChannel = new byte*[nbChannels];
            for (var i = 0; i < nbChannels; i++) pChannel[i] = (byte*)asioOutputBuffers[i];

            for (var i = 0; i < nbSamples; i++)
            {
                for (var ch = 0; ch < nbChannels; ch++)
                {
                    *pChannel[ch]++ = *inputSamples++;
                    *pChannel[ch]++ = *inputSamples++;
                    *pChannel[ch]++ = 0; // padding zero
                }
            }
        }
    }

    private static int ClampTo24Bit(double sampleValue)
    {
        sampleValue = sampleValue < -1.0 ? -1.0 : sampleValue > 1.0 ? 1.0 : sampleValue;
        return (int)(sampleValue * 8388607.0);
    }

    private static int ClampToInt(double sampleValue)
    {
        sampleValue = sampleValue < -1.0 ? -1.0 : sampleValue > 1.0 ? 1.0 : sampleValue;
        return (int)(sampleValue * 2147483647.0);
    }

    private static short ClampToShort(double sampleValue)
    {
        sampleValue = sampleValue < -1.0 ? -1.0 : sampleValue > 1.0 ? 1.0 : sampleValue;
        return (short)(sampleValue * 32767.0);
    }

    // BmpMod: ConvertedBytePerChannelMonoDsdCopy
    public static void ConvertedBytePerChannelMonoDsdCopy(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (byte*)inputInterleavedBuffer;
            // Use a trick (short instead of int to avoid any convertion from 16Bit to 32Bit)
            var outputChannel = (byte*)asioOutputBuffers[0];

            // Point to upper 16 bits of the 32Bits.
            //outputChannel++;
            for (var i = 0; i < nbSamples; i++)
            {
                *outputChannel = inputSamples[0];
                // Go to next cluster for source
                inputSamples += 1;
                // Go to next Byte for Output
                outputChannel += 1;
            }
        }
    }

    // BmpMod: ConvertedBytePerChannelStereoDsdCopy
    public static void ConvertedBytePerChannelStereoDsdCopy(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (byte*)inputInterleavedBuffer;
            // Use a trick (short instead of int to avoid any convertion from 16Bit to 32Bit)
            var leftSamples = (byte*)asioOutputBuffers[0];
            var rightSamples = (byte*)asioOutputBuffers[1];

            for (var i = 0; i < nbSamples; i++)
            {
                var value = inputSamples[0];
                *leftSamples = value;

                value = inputSamples[1];
                *rightSamples = value;

                // Go to next cluster for source
                inputSamples += 2;
                // Go to next Byte for Output
                leftSamples += 1;
                rightSamples += 1;
            }
        }
    }

    // BmpMod: ConvertedBytePerChannelMonoDsdRsv
    public static void ConvertedBytePerChannelMonoDsdRsv(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (byte*)inputInterleavedBuffer;
            // Use a trick (short instead of int to avoid any convertion from 16Bit to 32Bit)
            var outputChannel = (byte*)asioOutputBuffers[0];

            // Point to upper 16 bits of the 32Bits.
            //outputChannel++;
            for (var i = 0; i < nbSamples; i++)
            {
                *outputChannel = BitReverseTable[inputSamples[0]];
                // Go to next cluster for source
                inputSamples += 1;
                // Go to next Byte for Output
                outputChannel += 1;
            }
        }
    }

    // BmpMod: ConvertedBytePerChannelStereoDsdRsv
    public static void ConvertedBytePerChannelStereoDsdRsv(nint inputInterleavedBuffer, nint[] asioOutputBuffers, int nbChannels, int nbSamples)
    {
        unsafe
        {
            var inputSamples = (byte*)inputInterleavedBuffer;
            // Use a trick (short instead of int to avoid any convertion from 16Bit to 32Bit)
            var leftSamples = (byte*)asioOutputBuffers[0];
            var rightSamples = (byte*)asioOutputBuffers[1];

            for (var i = 0; i < nbSamples; i++)
            {
                var value = inputSamples[0];
                *leftSamples = BitReverseTable[value];

                value = inputSamples[1];
                *rightSamples = BitReverseTable[value];

                // Go to next cluster for source
                inputSamples += 2;
                // Go to next Byte for Output
                leftSamples += 1;
                rightSamples += 1;
            }
        }
    }

    private static readonly byte[] BitReverseTable =
    {
        0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
        0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
        0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
        0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
        0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
        0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
        0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
        0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
        0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
        0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
        0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
        0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
        0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
        0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
        0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
        0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
        0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
        0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
        0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
        0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
        0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
        0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
        0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
        0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
        0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
        0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
        0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
        0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
        0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
        0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
        0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
        0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff
    };
}
