﻿using Bmp.Core.Lite.Playback.Inputs;
using Bmp.Core.Lite.Playback.Outputs.NAudioASIO.Originals;
using NAudio.Wave;
using AsioAudioAvailableEventArgs = Bmp.Core.Lite.Playback.Outputs.NAudioASIO.Originals.AsioAudioAvailableEventArgs;

namespace Bmp.Core.Lite.Playback.Outputs.NAudioASIO;

/// <summary>
/// ASIO Out Player. New implementation using an internal C# binding.
///
/// This implementation is only supporting Short16Bit and Float32Bit formats and is optimized
/// for 2 outputs channels .
/// SampleRate is supported only if AsioDriver is supporting it
///
/// This implementation is probably the first AsioDriver binding fully implemented in C#!
///
/// Original Contributor: Mark Heath
/// New Contributor to C# binding : Alexandre Mutel - email: alexandre_mutel at yahoo.fr
/// </summary>
public class AsioOut : IWavePlayer
{
    public AsioDriverExt Driver { get; private set; }
    private IWaveProvider _sourceStream;
    private PlaybackState _playbackState;
    private int _nbSamples;
    private byte[] _waveBuffer;
    private AsioSampleConvertor.SampleConvertor _convertor;
    private string _driverName;

    private readonly SynchronizationContext _syncContext;
    private bool _isInitialized;
    private bool _nativeDsd;

    /// <summary>
    /// Playback Stopped
    /// </summary>
    public event EventHandler<StoppedEventArgs> PlaybackStopped;

    /// <summary>
    /// When recording, fires whenever recorded audio is available
    /// </summary>
    public event EventHandler<AsioAudioAvailableEventArgs>? AudioAvailable;

    /// <summary>
    /// Occurs when the driver settings are changed by the user, e.g. in the control panel.
    /// </summary>
    public event EventHandler DriverResetRequest;

    /// <summary>
    /// Initializes a new instance of the <see cref="AsioOut"/> class with the first
    /// available ASIO Driver.
    /// </summary>
    public AsioOut()
        : this(0)
    {
    }

    /// <summary>
    /// Initializes a new instance of the <see cref="AsioOut"/> class with the driver name.
    /// </summary>
    /// <param name="driverName">Name of the device.</param>
    public AsioOut(string driverName)
    {
        _syncContext = SynchronizationContext.Current;
        InitFromName(driverName);
    }

    /// <summary>
    /// Opens an ASIO output device
    /// </summary>
    /// <param name="driverIndex">Device number (zero based)</param>
    public AsioOut(int driverIndex)
    {
        _syncContext = SynchronizationContext.Current;
        var names = GetDriverNames();
        if (names.Length == 0)
        {
            throw new ArgumentException("There is no ASIO Driver installed on your system");
        }
        if (driverIndex < 0 || driverIndex > names.Length)
        {
            throw new ArgumentException($"Invalid device number. Must be in the range [0,{names.Length}]");
        }
        InitFromName(names[driverIndex]);
    }

    public void SetNativeDsd(bool enable = true)
    {
        try
        {
            if (enable)
            {
                Driver.Future_SetIoFormat_DSD();
                _nativeDsd = true;
            }
            else
            {
                Driver.Future_SetIoFormat_PCM();
                _nativeDsd = false;
            }
        }
        catch (Exception e)
        {
            throw new NotSupportedException("Not support native DSD", e);
        }
    }

    /// <summary>
    /// Releases unmanaged resources and performs other cleanup operations before the
    /// <see cref="AsioOut"/> is reclaimed by garbage collection.
    /// </summary>
    ~AsioOut()
    {
        Dispose();
    }

    /// <summary>
    /// Dispose
    /// </summary>
    public void Dispose()
    {
        if (Driver != null)
        {
            if (_playbackState != PlaybackState.Stopped)
            {
                Driver.Stop();
            }
            Driver.ResetRequestCallback = null;
            Driver.ReleaseDriver();
            Driver = null;
        }
    }

    /// <summary>
    /// Gets the names of the installed ASIO Driver.
    /// </summary>
    /// <returns>an array of driver names</returns>
    public static string[] GetDriverNames() => AsioDriver.GetAsioDriverNames();

    /// <summary>
    /// Determines whether ASIO is supported.
    /// </summary>
    /// <returns>
    ///     <c>true</c> if ASIO is supported; otherwise, <c>false</c>.
    /// </returns>
    public static bool IsSupported()
    {
        return GetDriverNames().Length > 0;
    }

    /// <summary>
    /// Determines whether this driver supports the specified sample rate.
    /// </summary>
    /// <param name="sampleRate">The samplerate to check.</param>
    /// <returns>
    ///   <c>true</c> if the specified sample rate is supported otherwise, <c>false</c>.
    /// </returns>
    public bool IsSampleRateSupported(int sampleRate)
    {
        return Driver.IsSampleRateSupported(sampleRate);
    }

    /// <summary>
    /// Inits the driver from the asio driver name.
    /// </summary>
    /// <param name="driverName">Name of the driver.</param>
    private void InitFromName(string driverName)
    {
        _driverName = driverName;

        // Instantiate the extended driver
        Driver = new AsioDriverExt(driverName);

        Driver.ResetRequestCallback = OnDriverResetRequest;
        ChannelOffset = 0;
    }

    private void OnDriverResetRequest()
    {
        DriverResetRequest?.Invoke(this, EventArgs.Empty);
    }

    /// <summary>
    /// Release driver
    /// </summary>
    private void ReleaseDriver(AsioDriver driver)
    {
        driver.DisposeBuffers();
        driver.ReleaseComAsioDriver();
    }

    /// <summary>
    /// Shows the control panel
    /// </summary>
    public void ShowControlPanel()
    {
        Driver.ShowControlPanel();
    }

    /// <summary>
    /// Starts playback
    /// </summary>
    public void Play()
    {
        if (_playbackState != PlaybackState.Playing)
        {
            _playbackState = PlaybackState.Playing;
            HasReachedEnd = false;
            Driver.Start();
        }
    }

    /// <summary>
    /// Stops playback
    /// </summary>
    public void Stop()
    {
        if (_playbackState == PlaybackState.Stopped) return;

        _playbackState = PlaybackState.Stopped;
        Driver?.Stop();
        HasReachedEnd = false;
        RaisePlaybackStopped(null);
    }

    /// <summary>
    /// Pauses playback
    /// </summary>
    public void Pause()
    {
        _playbackState = PlaybackState.Paused;
        Driver.Stop();
    }

    /// <summary>
    /// Initialises to play
    /// </summary>
    /// <param name="waveProvider">Source wave provider</param>
    public void Init(IWaveProvider waveProvider)
    {
        InitRecordAndPlayback(waveProvider, 0, -1);
    }

    /// <summary>
    /// Initialises to play, with optional recording
    /// </summary>
    /// <param name="waveProvider">Source wave provider - set to null for record only</param>
    /// <param name="recordChannels">Number of channels to record</param>
    /// <param name="recordOnlySampleRate">Specify sample rate here if only recording, ignored otherwise</param>
    public void InitRecordAndPlayback(IWaveProvider waveProvider, int recordChannels, int recordOnlySampleRate)
    {
        if (_isInitialized)
        {
            throw new InvalidOperationException("Already initialised this instance of AsioOut - dispose and create a new one");
        }

        _isInitialized = true;

        var desiredSampleRate = waveProvider != null ? waveProvider.WaveFormat.SampleRate : recordOnlySampleRate;

        // BmpMod: first to SetSampleRate then get asioSampleType

        if (!Driver.IsSampleRateSupported(desiredSampleRate))
        {
            throw new ArgumentException("SampleRate is not supported");
        }

        // ReSharper disable once CompareOfFloatsByEqualityOperator
        if (Driver.Capabilities.SampleRate != desiredSampleRate)
        {
            Driver.SetSampleRate(desiredSampleRate, _nativeDsd == false);
        }

        if (waveProvider != null)
        {
            _sourceStream = waveProvider;

            NumberOfOutputChannels = waveProvider.WaveFormat.Channels;

            // Select the correct sample convertor from WaveFormat -> ASIOFormat
            var asioSampleType = Driver.Capabilities.OutputChannelInfos[0].type;
            _convertor = AsioSampleConvertor.SelectSampleConvertor(waveProvider.WaveFormat, asioSampleType, (waveProvider as IHaveBitPerRawSample)?.BitPerRawSample, (waveProvider as IHaveAsioSampleFormat)?.AsioSampleFormat);

            switch (asioSampleType)
            {
                case AsioSampleType.DSDInt8LSB1:
                case AsioSampleType.DSDInt8MSB1:
                    OutputWaveFormat = _sourceStream.WaveFormat;
                    break;

                case AsioSampleType.Float32LSB:
                    OutputWaveFormat = WaveFormat.CreateIeeeFloatWaveFormat(waveProvider.WaveFormat.SampleRate, waveProvider.WaveFormat.Channels);
                    break;

                case AsioSampleType.Int32LSB:
                    OutputWaveFormat = new WaveFormat(waveProvider.WaveFormat.SampleRate, 32, waveProvider.WaveFormat.Channels);
                    break;

                case AsioSampleType.Int16LSB:
                    OutputWaveFormat = new WaveFormat(waveProvider.WaveFormat.SampleRate, 16, waveProvider.WaveFormat.Channels);
                    break;

                case AsioSampleType.Int24LSB:
                    OutputWaveFormat = new WaveFormat(waveProvider.WaveFormat.SampleRate, 24, waveProvider.WaveFormat.Channels);
                    break;

                default:
                    throw new NotSupportedException($"{asioSampleType} not currently supported");
            }
        }
        else
        {
            NumberOfOutputChannels = 0;
        }

        // Plug the callback
        Driver.FillBufferCallback = driver_BufferUpdate;

        NumberOfInputChannels = recordChannels;
        // Used Prefered size of ASIO Buffer
        _nbSamples = Driver.CreateBuffers(NumberOfOutputChannels, NumberOfInputChannels, false);
        if (_nativeDsd) _nbSamples /= 8;
        Driver.SetChannelOffset(ChannelOffset, InputChannelOffset); // will throw an exception if channel offset is too high

        if (waveProvider != null)
        {
            // make a buffer big enough to read enough from the sourceStream to fill the ASIO buffers

            _waveBuffer = _nativeDsd
                ? new byte[_nbSamples * NumberOfOutputChannels]
                : new byte[_nbSamples * NumberOfOutputChannels * waveProvider.WaveFormat.BitsPerSample / 8];
        }
    }

    /// <summary>
    /// driver buffer update callback to fill the wave buffer.
    /// </summary>
    /// <param name="inputChannels">The input channels.</param>
    /// <param name="outputChannels">The output channels.</param>
    private void driver_BufferUpdate(nint[] inputChannels, nint[] outputChannels)
    {
        if (NumberOfInputChannels > 0)
        {
            var audioAvailable = AudioAvailable;
            if (audioAvailable != null)
            {
                var args = new AsioAudioAvailableEventArgs(inputChannels, outputChannels, _nbSamples,
                    Driver.Capabilities.InputChannelInfos[0].type);
                audioAvailable(this, args);
                if (args.WrittenToOutputBuffers)
                    return;
            }
        }

        if (NumberOfOutputChannels > 0)
        {
            var read = _sourceStream.Read(_waveBuffer, 0, _waveBuffer.Length);
            if (read < _waveBuffer.Length)
            {
                // we have reached the end of the input data - clear out the end
                Array.Clear(_waveBuffer, read, _waveBuffer.Length - read);
            }

            // Call the convertor
            unsafe
            {
                // TODO : check if it's better to lock the buffer at initialization?
                fixed (void* pBuffer = &_waveBuffer[0])
                {
                    _convertor(new nint(pBuffer), outputChannels, NumberOfOutputChannels, _nbSamples);
                }
            }

            if (read == 0)
            {
                HasReachedEnd = true;
            }
        }
    }

    /// <summary>
    /// Gets the latency (in samples) of the playback driver
    /// </summary>
    public int PlaybackLatency
    {
        get
        {
            Driver.Driver.GetLatencies(out var temp, out var latency);
            return latency;
        }
    }

    /// <summary>
    /// A flag to let you know that we have reached the end of the input file
    /// Useful if AutoStop is set to false
    /// You can monitor this yourself and call Stop when it is true
    /// </summary>
    public bool HasReachedEnd { get; private set; }

    /// <summary>
    /// Playback State
    /// </summary>
    public PlaybackState PlaybackState => _playbackState;

    /// <summary>
    /// Driver Name
    /// </summary>
    public string DriverName => _driverName;

    /// <summary>
    /// The number of output channels we are currently using for playback
    /// (Must be less than or equal to DriverOutputChannelCount)
    /// </summary>
    public int NumberOfOutputChannels { get; private set; }

    /// <summary>
    /// The number of input channels we are currently recording from
    /// (Must be less than or equal to DriverInputChannelCount)
    /// </summary>
    public int NumberOfInputChannels { get; private set; }

    /// <summary>
    /// The maximum number of input channels this ASIO driver supports
    /// </summary>
    public int DriverInputChannelCount => Driver.Capabilities.NbInputChannels;

    /// <summary>
    /// The maximum number of output channels this ASIO driver supports
    /// </summary>
    public int DriverOutputChannelCount => Driver.Capabilities.NbOutputChannels;

    /// <summary>
    /// The number of samples per channel, per buffer.
    /// </summary>
    public int FramesPerBuffer
    {
        get
        {
            if (!_isInitialized)
                throw new Exception("Not initialized yet. Call this after calling Init");

            return _nbSamples;
        }
    }

    /// <summary>
    /// By default the first channel on the input WaveProvider is sent to the first ASIO output.
    /// This option sends it to the specified channel number.
    /// Warning: make sure you don't set it higher than the number of available output channels -
    /// the number of source channels.
    /// n.b. Future NAudio may modify this
    /// </summary>
    public int ChannelOffset { get; set; }

    /// <summary>
    /// Input channel offset (used when recording), allowing you to choose to record from just one
    /// specific input rather than them all
    /// </summary>
    public int InputChannelOffset { get; set; }

    /// <summary>
    /// Sets the volume (1.0 is unity gain)
    /// Not supported for ASIO Out. Set the volume on the input stream instead
    /// </summary>
    [Obsolete("this function will be removed in a future NAudio as ASIO does not support setting the volume on the device")]
    public float Volume
    {
        get
        {
            return 1.0f;
        }
        set
        {
            if (value != 1.0f)
            {
                throw new InvalidOperationException("AsioOut does not support setting the device volume");
            }
        }
    }

    /// <inheritdoc/>
    public WaveFormat OutputWaveFormat { get; private set; }

    private void RaisePlaybackStopped(Exception e)
    {
        var handler = PlaybackStopped;
        if (handler != null)
        {
            if (_syncContext == null)
            {
                handler(this, new StoppedEventArgs(e));
            }
            else
            {
                _syncContext.Post(state => handler(this, new StoppedEventArgs(e)), null);
            }
        }
    }

    /// <summary>
    /// Get the input channel name
    /// </summary>
    /// <param name="channel">channel index (zero based)</param>
    /// <returns>channel name</returns>
    public string AsioInputChannelName(int channel)
    {
        return channel > DriverInputChannelCount ? "" : Driver.Capabilities.InputChannelInfos[channel].name;
    }

    /// <summary>
    /// Get the output channel name
    /// </summary>
    /// <param name="channel">channel index (zero based)</param>
    /// <returns>channel name</returns>
    public string AsioOutputChannelName(int channel)
    {
        return channel > DriverOutputChannelCount ? "" : Driver.Capabilities.OutputChannelInfos[channel].name;
    }
}
